import logging
import threading
from typing import List, Dict, Tuple, Union, Any

from sfdevtools.data_cache.DCache import DCache
from sfdevtools.data_cache.DComponents import StrategyInsight, TS_Order, TS_Trade
import sfdevtools.grpc_protos.ts_cop_pb2 as ts_cop_pb2
import sfdevtools.grpc_protos.ts_cop_pb2_grpc as ts_cop_pb2_grpc
import sfdevtools.devTools.TimelyCache as TimelyCache

class DCacheManager(object):
    def __init__(self):
        self.__logger: logging.Logger = None
        self.__dcache: Dict[str, DCache] = dict() # key: cache name, value: dcache
        self.__mutex: threading.Lock = threading.Lock()

    def init_component(self, logger: logging.Logger) -> None:
        self.__logger = logger

    def create_cache(self
                     , cache_name: str
                     , max_hist_orders: int = 1000
                     , max_hist_trades: int = 1000
                     , max_hist_si: int = 1000) -> DCache:
        with self.__mutex:
            if cache_name not in self.__dcache:
                self.__logger.info(f"Create DCache with name: {cache_name}")
                new_dcache = DCache()
                self.__dcache[cache_name] = new_dcache
                new_dcache.init_component(logger=self.__logger
                                          , max_hist_orders=max_hist_orders
                                          , max_hist_si=max_hist_si
                                          , max_hist_trades=max_hist_trades)
                return new_dcache
            else:
                return self.__dcache[cache_name]

    def get_cache(self, cache_name: str) -> Union[bool, DCache]:
        with self.__mutex:
            if cache_name in self.__dcache:
                return (True, self.__dcache[cache_name])
            else:
                return (False, None)

    def get_strategy_insight(self, cache_name: str) -> Union[bool, List]:
        ret, dcache = self.get_cache(cache_name=cache_name)
        if not ret:
            return (False, None)
        return (True, dcache.get_strategy_insight())

    def save_si(self
                , cache_name: str
                , si: List[StrategyInsight]) -> Union[bool, bool]:
        dcache = self.create_cache(cache_name=cache_name)
        return (True, dcache.save_si(si=si))

    def is_latest_si(self, cache_name: str, si: List[StrategyInsight]) -> Union[bool, bool]:
        ret, dcache = self.get_cache(cache_name=cache_name)
        if not ret:
            return (False, None)
        return (True, dcache.is_latest_si(si=si))

    def is_si_exist(self, cache_name: str, si_id: str) -> Union[bool, bool]:
        ret, dcache = self.get_cache(cache_name=cache_name)
        if not ret:
            return (False, None)
        return (True, dcache.is_si_exist(si_id=si_id))

    def get_ci_id(self, cache_name: str) -> Union[bool, str]:
        ret, dcache = self.get_cache(cache_name=cache_name)
        if not ret:
            return (False, None)
        return (True, dcache.get_ci_id())

    def is_order_exist(self, cache_name: str, platform_order_id: str) -> Union[bool, bool]:
        ret, dcache = self.get_cache(cache_name=cache_name)
        if not ret:
            return (False, None)
        return (True, dcache.is_order_exist(platform_order_id=platform_order_id))

    def get_trade(self, cache_name: str, trade_id: str) -> Union[bool, Union[bool, TS_Trade]]:
        ret, dcache = self.get_cache(cache_name=cache_name)
        if not ret:
            return (False, (False, None))
        return (True, dcache.get_trade(trade_id=trade_id))

    def get_order(self, cache_name: str, platform_order_id: str) -> Union[bool, Union[bool, TS_Order]]:
        ret, dcache = self.get_cache(cache_name=cache_name)
        if not ret:
            return (False, (False, None))
        return (True, dcache.get_order(platform_order_id=platform_order_id))

    def save_ci(self, cache_name: str, ci: Dict[str, Any]) -> bool:
        dcache = self.create_cache(cache_name=cache_name)
        dcache.save_ci(ci=ci)
        return True

    def save_ci_id(self, cache_name: str, ci_id: str) -> bool:
        dcache = self.create_cache(cache_name=cache_name)
        dcache.save_ci_id(ci_id=ci_id)
        return True

    def save_order(self, cache_name: str, order: TS_Order) -> bool:
        dcache = self.create_cache(cache_name=cache_name)
        dcache.save_order(order=order)
        return True

    def save_orders(self, cache_name: str, orders: List[TS_Order]) -> bool:
        dcache = self.create_cache(cache_name=cache_name)
        dcache.save_orders(orders=orders)
        return True

    def is_trade_exist(self, cache_name: str, trade_id: str) -> Union[bool, bool]:
        ret, dcache = self.get_cache(cache_name=cache_name)
        if not ret:
            return (False, None)
        return (True, dcache.is_trade_exist(trade_id=trade_id))

    def save_trade(self, cache_name: str, trade: TS_Trade) -> None:
        dcache = self.create_cache(cache_name=cache_name)
        dcache.save_trade(trade=trade)
        return None

    def save_trades(self, cache_name: str, trades: List[TS_Trade]) -> None:
        dcache = self.create_cache(cache_name=cache_name)
        dcache.save_trades(trades=trades)
        return None

    def save_fid(self
                 , cache_name: str
                 , msg_id: str
                 , msg_type: ts_cop_pb2.Cop.MsgType
                 , fid_num: int
                 , fid_value: Any) -> None:
        dcache = self.create_cache(cache_name=cache_name)
        dcache.save_fid(msg_id=msg_id, msg_type=msg_type, fid_num=fid_num, fid_value=fid_value)

    def get_fid(self
                , cache_name: str
                , msg_id: str
                , msg_type: ts_cop_pb2.Cop.MsgType
                , fid_num: int) -> Union[bool, Union[bool, Union[bool, Any]]]:
        ret, dcache = self.get_cache(cache_name=cache_name)
        if not ret:
            return (False, (False, None))
        return (True, dcache.get_fid(msg_id=msg_id
                                     , msg_type=msg_type
                                     , fid_num=fid_num))
    def get_order_snapshot_cache(self
                                 , cache_name: str) -> TimelyCache.TimelyCache_DupKey:
        dcache = self.create_cache(cache_name=cache_name)
        return dcache.get_order_snapshot_cache()

    def get_order_hist_cache(self
                             , cache_name: str) -> TimelyCache.TimelyCache_UniKey:
        dcache = self.create_cache(cache_name=cache_name)
        return dcache.get_order_hist_cache()



import logging
import pymongo
from pymongo import MongoClient

class MongoDBCtrl(object):
    def __init__(self
                 , db_user: str
                 , db_pw: str
                 , db_host: str
                 , db_port: str
                 , db_name: str
                 , logger: logging.Logger
                 , db_conn_protocol: str = "mongodb+srv"
                 , connection_time_ms: int = 20000):
        self._logger: logging.Logger = logger
        self._db_client = None
        if "mongodb" == db_conn_protocol:
            self._db_client = MongoClient(f"mongodb://{db_user}:{db_pw}@{db_host}:{db_port}/?authSource={db_name}"
                                          , serverSelectionTimeoutMS=connection_time_ms)
        else:
            self._db_client = MongoClient(f"mongodb+srv://{db_user}:{db_pw}@{db_host}/?authSource={db_name}"
                                          , serverSelectionTimeoutMS=connection_time_ms)

        self._db = self._db_client[db_name]

    def get_db(self, db_name: str = None) -> pymongo.database.Database:
        if None is db_name:
            return self._db
        else:
            return self._db_client[db_name]

    def get_collection(self, collection_name: str) -> pymongo.collection.Collection:
        return self._db[collection_name]



from sqlalchemy import create_engine, MetaData # poetry add psycopg2-binary, sqlalchemy
from sqlalchemy.ext.automap import automap_base
import sqlalchemy
from sqlalchemy.orm import Session
from typing import List, Dict, Tuple, Any
import pandas as pd
import logging

# https://hackersandslackers.com/database-queries-sqlalchemy-orm/
class PostgresDBCtrl(object):
    def __init__(self
                 , db_user: str
                 , db_pw: str
                 , db_host: str
                 , db_port: str
                 , db_name: str
                 , db_schema: str
                 , logger: logging.Logger
                 , include_tables: List[str] = None
                 , connection_time: int = 10):
        self._logger = logger

        self._logger.info(f"db config db_user: {db_user}, db_host: {db_host}, db_port: {db_port}, db_name: {db_name}")
        # prepare database connection
        self._engine = create_engine(f"postgresql+psycopg2://{db_user}:{db_pw}@{db_host}:{db_port}/{db_name}"
                                     , connect_args={"connect_timeout": connection_time})

        # prepare metadata, metadata will contain tables information
        self._metadata = MetaData()
        if None == include_tables:
            self._metadata.reflect(bind=self._engine)
        else:
            self._metadata.reflect(bind=self._engine, only=include_tables)

        # prepare automap base, this will map tables to python class
        self._automap_base = automap_base(metadata=self._metadata)
        self._automap_base.prepare(self._engine, reflect=True, schema=db_schema)

        # successfully connected
        self._logger.info("Successfully connected")
        self._logger.info(self._metadata.tables.keys())
        self._logger.info(self._automap_base.classes.keys())

    def get_classes(self) -> sqlalchemy.util._collections.Properties:
        return self._automap_base.classes

    def get_tables(self) -> sqlalchemy.util._collections.FacadeDict:
        return self._metadata.tables

    def get_session(self) -> sqlalchemy.orm.session.Session:
        return Session(self._engine)

    def get_connection(self) -> sqlalchemy.engine.base.Connection:
        return self._engine.connect()

    def get_query(self, *args) -> sqlalchemy.orm.query.Query:
        return Session(self._engine).query(*args)

    def get_df_from_query(self, query: sqlalchemy.orm.query.Query, is_debug: bool = False) -> pd.DataFrame:
        if is_debug:
            self._logger.debug(query.statement.compile(compile_kwargs={"literal_binds": True}))
        return pd.read_sql(sql=query.statement, con=self.get_connection())

    def add_records(self, db_records: List[Any]) -> bool:
        db_session = Session(self._engine)
        db_session.add_all(db_records)

        try:
            db_session.commit()
            return True
        except SQLAlchemyError as e:
            db_session.rollback()
            return False
        finally:
            db_session.close()

    def upsert_records(self, db_records: List[Any]) -> bool:
        if len(db_records) == 0:
            return True

        db_session = Session(self._engine)
        for db_record in db_records:
            db_session.merge(db_record)

        try:
            db_session.commit()
            return True
        except SQLAlchemyError as e:
            db_session.rollback()
            return False
        finally:
            db_session.close()



import logging
import zmq
import threading
import signal

import sfdevtools.grpc_protos.ts_cop_pb2 as ts_cop_pb2
import sfdevtools.grpc_protos.ts_cop_pb2_grpc as ts_cop_pb2_grpc

class TSDataCtrl(object):
    def __init__(self):
        self.__logger: logging.Logger = None
        self.__pub_mutex: threading.Lock = threading.Lock()
        self.__sub_mutex: threading.Lock = threading.Lock()
        self.__publisher: zmq.sugar.socket.Socket = None
        self.__subscriber: zmq.sugar.socket.Socket = None
        self.__sender_name: str = ""
        self.__instance_id: int = 0
        self.__seq_num: int = 0
        self.__cb_fun = None
        self.__sub_main_thread: threading.Thread = None
        self.__is_subscription_run: bool = False
        self.__sub_timeout_ms: int = 100

    def init_component(self
                       , logger: logging.Logger):
        self.__logger = logger

    def init_publish(self
                     , pub_host: str
                     , pub_port: str
                     , sender_name: str
                     , instance_id: int):
        ctx = zmq.Context.instance()
        self.__publisher = ctx.socket(zmq.PUB)
        publisher_url = f"tcp://{pub_host}:{pub_port}"
        self.__logger.info(f"publisher_url: {publisher_url}")
        self.__publisher.connect(publisher_url)
        self.__logger.info("socket connect done")

        self.__sender_name = sender_name
        self.__instance_id = instance_id

    def init_subscribe(self
                       , sub_host: str
                       , sub_port: str
                       , timeout_ms: int = 100):
        ctx = zmq.Context.instance()
        self.__subscriber = ctx.socket(zmq.SUB)
        subscriber_url = f"tcp://{sub_host}:{sub_port}"
        self.__logger.info(f"subscriber_url: {subscriber_url}")
        self.__subscriber.connect(subscriber_url)
        self.__logger.info("socket connect done")

        self.__sub_timeout_ms = timeout_ms

    def reg_sub_topic(self, topic: str):
        self.__logger.info(f"Subscribe topic: {topic}")
        self.__subscriber.setsockopt(zmq.SUBSCRIBE, str.encode(topic))

    def start_subscription(self, mode: str = "simple"):
        with self.__sub_mutex:
            self.__is_subscription_run = True
            if mode == "simple":
                self.__sub_main_thread = threading.Thread(target=self.sub_main_simple)
            else:
                self.__sub_main_thread = threading.Thread(target=self.sub_main_extract)
            self.__sub_main_thread.start()

            signal.signal(signal.SIGINT, self.__cleanup)

    def stop_subscription(self):
        with self.__sub_mutex:
            self.__is_subscription_run = False

        self.__sub_main_thread.join()

    def reg_callback(self, cb_fun):
        with self.__sub_mutex:
            self.__cb_fun = cb_fun

    def unreg_callback(self):
        with self.__sub_mutex:
            self.__cb_fun = None

    def send_cop(self, msg_id: str, msg_type: ts_cop_pb2.Cop.MsgType, topic: str, cop: ts_cop_pb2.Cop):
        with self.__pub_mutex:
            cop.msg_id = msg_id
            cop.msg_type = msg_type
            cop.sender = self.__sender_name
            cop.instance_id = self.__instance_id
            cop.seq_num = self.__seq_num
            self.__seq_num += 1
            self.__publisher.send_multipart([str.encode(topic), cop.SerializeToString()])

    def reset_msg_seq_num(self):
        with self.__pub_mutex:
            self.__seq_num = 0

    def sub_main_extract(self):
        self.__logger.info("Start")

        poller = zmq.Poller()
        poller.register(self.__subscriber, zmq.POLLIN)
        while self.__is_subscription():
            sockets = dict(poller.poll(self.__sub_timeout_ms))

            if not self.__is_subscription():
                break

            # get and process data
            if self.__subscriber in sockets:
                topic_encoded, data = self.__subscriber.recv_multipart()

                with self.__sub_mutex:
                    if self.__cb_fun is not None:
                        cop = ts_cop_pb2.Cop()
                        cop.ParseFromString(data)
                        topic = topic_encoded.decode("utf-8")
                        # int32
                        for fid_num, fid_value in cop.int32_map.items():
                            self.__cb_fun(topic=topic, cop=cop, fid_num=fid_num, fid_value=fid_value)
                        # int64
                        for fid_num, fid_value in cop.int64_map.items():
                            self.__cb_fun(topic=topic, cop=cop, fid_num=fid_num, fid_value=fid_value)
                        # float
                        for fid_num, fid_value in cop.float_map.items():
                            self.__cb_fun(topic=topic, cop=cop, fid_num=fid_num, fid_value=fid_value)
                        # double
                        for fid_num, fid_value in cop.double_map.items():
                            self.__cb_fun(topic=topic, cop=cop, fid_num=fid_num, fid_value=fid_value)
                        # string
                        for fid_num, fid_value in cop.string_map.items():
                            self.__cb_fun(topic=topic, cop=cop, fid_num=fid_num, fid_value=fid_value)
                        # bool
                        for fid_num, fid_value in cop.bool_map.items():
                            self.__cb_fun(topic=topic, cop=cop, fid_num=fid_num, fid_value=fid_value)
                        # order
                        for fid_num, fid_value in cop.order_map.items():
                            self.__cb_fun(topic=topic, cop=cop, fid_num=fid_num, fid_value=fid_value.order_list)
                        # trade
                        for fid_num, fid_value in cop.trade_map.items():
                            self.__cb_fun(topic=topic, cop=cop, fid_num=fid_num, fid_value=fid_value.trade_list)
                        # si
                        for fid_num, fid_value in cop.si_map.items():
                            self.__cb_fun(topic=topic, cop=cop, fid_num=fid_num, fid_value=fid_value.si_list)
                        # ci
                        for fid_num, fid_value in cop.ci_map.items():
                            for ci in fid_value.ci_list:
                                self.__cb_fun(topic=topic, cop=cop, fid_num=fid_num, fid_value=ci.value)

        self.__logger.info("End")

    def sub_main_simple(self):
        self.__logger.info("Start")

        poller = zmq.Poller()
        poller.register(self.__subscriber, zmq.POLLIN)
        while self.__is_subscription():
            sockets = dict(poller.poll(self.__sub_timeout_ms))

            if not self.__is_subscription():
                break

            # get and process data
            if self.__subscriber in sockets:
                topic, data = self.__subscriber.recv_multipart()

                with self.__sub_mutex:
                    if self.__cb_fun is not None:
                        cop = ts_cop_pb2.Cop()
                        cop.ParseFromString(data)
                        self.__cb_fun(topic.decode("utf-8"), cop)

            if not self.__is_subscription():
                break

        self.__logger.info("End")

    def __cleanup(self, signum, frame):
        self.stop_subscription()
        exit(0)

    def __is_subscription(self):
        with self.__sub_mutex:
            return self.__is_subscription_run
